# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kwil/tx/v1/ping.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15kwil/tx/v1/ping.proto\x12\x02tx\"\x1e\n\x0bPingRequest\x12\x0f\n\x07message\x18\x01 \x01(\t\"\x1f\n\x0cPingResponse\x12\x0f\n\x07message\x18\x01 \x01(\tB\x1eZ\x1ckwil/api/protobuf/tx/v1;txpbb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'kwil.tx.v1.ping_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\034kwil/api/protobuf/tx/v1;txpb'
  _PINGREQUEST._serialized_start=29
  _PINGREQUEST._serialized_end=59
  _PINGRESPONSE._serialized_start=61
  _PINGRESPONSE._serialized_end=92
# @@protoc_insertion_point(module_scope)



import hashlib
from eth_account.signers.local import LocalAccount

from kwil.types import TxParams, TxPayloadType, Nonce
from kwil._utils.signature import sign


# generate_tx_hash generates a hash of the transaction
# it does this by hashing the payload type, payload, fee, and nonce
def generate_tx_hash(
    payload: bytes, payload_type: TxPayloadType, fee: str, nonce: Nonce
) -> bytes:
    data = bytearray()

    # convert payload type to bytes
    payload_type_bytes = payload_type.to_bytes(4, byteorder="little")
    data.extend(payload_type_bytes)

    # hash payload
    payload_hash = hashlib.sha384(payload).digest()
    data.extend(payload_hash)

    # add fee
    data.extend(fee.encode())

    # convert nonce to bytes
    nonce_bytes = nonce.to_bytes(8, byteorder="little")
    data.extend(nonce_bytes)

    return hashlib.sha384(data).digest()


def sign_tx(params: TxParams, wallet: LocalAccount) -> TxParams:
    tx_hash = generate_tx_hash(
        params["payload"],
        params["payloadType"],
        params["fee"],
        params["nonce"],
    )

    params["hash"] = tx_hash
    params["signature"] = sign(tx_hash, wallet.key)
    params["sender"] = wallet.address
    return params



from typing import Optional, Sequence, Type, Union, Callable, Any

import grpc

from .base import BaseProvider
from .grpc import GRPCProvider
from kwil.types import RPCEndpoint, RPCResponse
from kwil.exceptions import CannotHandleRequest


class AutoProvider(BaseProvider):
    default_providers = (GRPCProvider,)

    _active_provider = None

    def __init__(
        self,
        # either a callable that returns a provider or a provider class
        potential_providers: Optional[
            Sequence[Union[Type[BaseProvider], Callable[..., BaseProvider]]]
        ] = None,
    ) -> None:
        if potential_providers:
            self._potential_providers = potential_providers
        else:
            self._potential_providers = self.default_providers

    def make_request(self, method: RPCEndpoint, params: Any) -> RPCResponse:
        try:
            return self._proxy_request(method, params)
        except grpc.RpcError:
            return self._proxy_request(method, params, use_cache=False)

    def is_connected(self) -> bool:
        provider = self._get_active_provider(use_cache=True)
        return provider is not None and provider.is_connected()

    def _proxy_request(
        self, method: RPCEndpoint, params: Any, use_cache: bool = True
    ) -> RPCResponse:
        provider = self._get_active_provider(use_cache)
        if provider is None:
            raise CannotHandleRequest(
                "Could not discover provider while making request: "
                "method:{0}\n"
                "params:{1}\n".format(method, params)
            )

        return provider.make_request(method, params)

    def _get_active_provider(self, use_cache: bool) -> Optional[BaseProvider]:
        if use_cache and self._active_provider is not None:
            return self._active_provider

        for Provider in self._potential_providers:
            provider = Provider()
            if provider is not None and provider.is_connected():
                self._active_provider = provider
                return provider

        return None



import os
from collections import OrderedDict
import threading
import logging
from typing import Any, Dict, List

from kwil._utils.naming_converter import camel_to_snake, rpc_to_grpc_method_name
from kwil._utils.grpc import Client as GRPCClient
from kwil.types import URI

DEFAULT_CACHE_SIZE = 20
DEFAULT_TIMEOUT = 10
logger = logging.getLogger(__name__)


class ConnectionCache:
    def __init__(self, size: int):
        self.size = size
        self._connections: OrderedDict[str, Any] = OrderedDict()

    def cache(self, key: str, value: Any) -> Dict[str, Any]:
        evicted_items = None

        if key not in self._connections:
            while len(self._connections) >= self.size:
                if evicted_items is None:
                    evicted_items = {}
                # FIFO
                k, v = self._connections.popitem(last=False)
                evicted_items[k] = v
        self._connections[key] = value
        return evicted_items

    def get_cache_entry(self, key: str) -> Any:
        return self._connections[key]

    def clear(self) -> None:
        self._connections.clear()

    def __contains__(self, item: str) -> bool:
        return item in self._connections

    def __len__(self) -> int:
        return len(self._connections)


_conn_cache = ConnectionCache(size=DEFAULT_CACHE_SIZE)
_conn_cache_lock = threading.Lock()


def _get_conn(endpoint_uri: str) -> Any:
    cache_key = f"grpc:{threading.get_ident()}:{endpoint_uri}"

    if cache_key in _conn_cache:
        return _conn_cache.get_cache_entry(cache_key)

    # cache new connection
    evicted_items = None
    with _conn_cache_lock:
        if cache_key not in _conn_cache:
            evicted_items = _conn_cache.cache(cache_key, GRPCClient(endpoint_uri))
            logger.debug("connection cached: %s", endpoint_uri)

    if evicted_items is not None:
        evicted_conns = evicted_items.values()
        for conn in evicted_conns:
            logger.debug("Cache full. Evicted conn: %s", conn)

        # close evicted connections after timeout in separate thread
        threading.Timer(
            DEFAULT_TIMEOUT + 0.1, _close_evicted_conns, args=[evicted_conns]
        )

    return _conn_cache.get_cache_entry(cache_key)


def _close_evicted_conns(conns: List[GRPCClient]) -> None:
    for conn in conns:
        conn.close()
        logger.debug("Evicted conn closed: %s", conn)


def grpc_request(endpoint_uri, method, args):
    conn = _get_conn(endpoint_uri)
    method = rpc_to_grpc_method_name(method)
    method = camel_to_snake(method)
    grpc_method = getattr(conn, method)
    # catch exceptions here
    resp = grpc_method(*args)
    return resp


def get_default_grpc_endpoint() -> URI:
    return URI(os.environ.get("KWIL_GRPC_PROVIDER_URI", "localhost:50051"))



from typing import Any, Dict, Callable

from toolz import compose, identity
import eth_utils as etu

from kwil.exceptions import InvalidAddress
from kwil._utils.utils import apply_validator_at_index
from kwil._utils.rpcs import GRPC
from kwil.types import RPCEndpoint


def is_not_address_string(value: Any) -> bool:
    return (
        etu.is_string(value)
        and not etu.is_bytes(value)
        and not etu.is_checksum_address(value)
        and not etu.is_hex_address(value)
    )


def validate_query(value: Any) -> None:
    """
    Helper function for validating a query.
    """
    if not isinstance(value, str):
        raise TypeError("Query {} must be provided as a string".format(value))


def validate_db_identifier(value: Any) -> None:
    """
    Helper function for validating a database identifier.
    """
    if not isinstance(value, str):
        raise TypeError(
            "Database identifier {} must be provided as a string".format(value)
        )

    if not value.startswith("x"):
        raise TypeError('Database identifier {} must start with "x"'.format(value))

    if not len(value) == 57:  # sha224 is 28 bytes, so 56 hex chars + 1 prefix
        raise TypeError("Database identifier {} must be 29 bytes".format(value))


def validate_address(value: Any) -> None:
    """
    Helper function for validating an address. Here we use eth address validation.
    """
    if not isinstance(value, str):
        raise TypeError("Address {} must be provided as a string".format(value))

    if not etu.is_hex_address(value):
        raise InvalidAddress(
            "Address must be 20 bytes, as a hex string with a 0x prefix", value
        )

    if not etu.is_checksum_address(value):
        raise InvalidAddress("Address has a invalid checksum", value)


def validate_tx_params(value: Any) -> None:
    """
    Helper function for validating transaction parameters.
    """
    if not isinstance(value, dict):
        raise TypeError("Transaction parameters must be provided as a dictionary")

    if "sender" in value:
        validate_address(value["sender"])

    if "fee" in value and not isinstance(value["fee"], str):
        raise TypeError("Fee must be provided as an integer")

    if "nonce" in value and not isinstance(value["nonce"], int):
        raise TypeError("Nonce must be provided as an integer")

    if "payloadType" in value and not isinstance(value["payloadType"], int):
        raise TypeError("PayloadType must be provided as a integer")

    if "payload" in value and not isinstance(value["payload"], bytes):
        raise TypeError("Payload must be provided as a string")

    if "hash" in value and not isinstance(value["hash"], bytes):
        raise TypeError("Hash must be provided as an bytes")

    if "signature" in value and not isinstance(value["signature"], dict):
        raise TypeError("Signature must be provided as an bytes")


request_validators: Dict[RPCEndpoint, Callable[..., Any]] = {
    GRPC.kwil_ping: identity,
    GRPC.kwil_getConfig: identity,
    GRPC.kwil_getSchema: apply_validator_at_index(validate_db_identifier, 0),
    GRPC.kwil_getAccount: apply_validator_at_index(validate_address, 0),
    GRPC.kwil_listDatabase: apply_validator_at_index(validate_address, 0),
    GRPC.kwil_broadcast: apply_validator_at_index(validate_tx_params, 0),
    GRPC.kwil_estimatePrice: apply_validator_at_index(validate_tx_params, 0),
    GRPC.kwil_query: compose(
        apply_validator_at_index(validate_db_identifier, 0),
        apply_validator_at_index(validate_query, 1),
    ),
}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kwil/tx/v1/service.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from kwil.tx.v1 import broadcast_pb2 as kwil_dot_tx_dot_v1_dot_broadcast__pb2
from kwil.tx.v1 import price_pb2 as kwil_dot_tx_dot_v1_dot_price__pb2
from kwil.tx.v1 import query_pb2 as kwil_dot_tx_dot_v1_dot_query__pb2
from kwil.tx.v1 import ping_pb2 as kwil_dot_tx_dot_v1_dot_ping__pb2
from kwil.tx.v1 import account_pb2 as kwil_dot_tx_dot_v1_dot_account__pb2
from kwil.tx.v1 import config_pb2 as kwil_dot_tx_dot_v1_dot_config__pb2
from kwil.tx.v1 import list_pb2 as kwil_dot_tx_dot_v1_dot_list__pb2
from kwil.tx.v1 import dataset_pb2 as kwil_dot_tx_dot_v1_dot_dataset__pb2
from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18kwil/tx/v1/service.proto\x12\x02tx\x1a\x1akwil/tx/v1/broadcast.proto\x1a\x16kwil/tx/v1/price.proto\x1a\x16kwil/tx/v1/query.proto\x1a\x15kwil/tx/v1/ping.proto\x1a\x18kwil/tx/v1/account.proto\x1a\x17kwil/tx/v1/config.proto\x1a\x15kwil/tx/v1/list.proto\x1a\x18kwil/tx/v1/dataset.proto\x1a\x1cgoogle/api/annotations.proto2\xcc\x05\n\tTxService\x12V\n\tBroadcast\x12\x14.tx.BroadcastRequest\x1a\x15.tx.BroadcastResponse\"\x1c\x82\xd3\xe4\x93\x02\x16\"\x11/api/v1/broadcast:\x01*\x12g\n\rEstimatePrice\x12\x18.tx.EstimatePriceRequest\x1a\x19.tx.EstimatePriceResponse\"!\x82\xd3\xe4\x93\x02\x1b\"\x16/api/v1/estimate_price:\x01*\x12\x46\n\x05Query\x12\x10.tx.QueryRequest\x1a\x11.tx.QueryResponse\"\x18\x82\xd3\xe4\x93\x02\x12\"\r/api/v1/query:\x01*\x12_\n\nGetAccount\x12\x15.tx.GetAccountRequest\x1a\x16.tx.GetAccountResponse\"\"\x82\xd3\xe4\x93\x02\x1c\x12\x1a/api/v1/accounts/{address}\x12?\n\x04Ping\x12\x0f.tx.PingRequest\x1a\x10.tx.PingResponse\"\x14\x82\xd3\xe4\x93\x02\x0e\x12\x0c/api/v1/ping\x12P\n\tGetConfig\x12\x14.tx.GetConfigRequest\x1a\x15.tx.GetConfigResponse\"\x16\x82\xd3\xe4\x93\x02\x10\x12\x0e/api/v1/config\x12_\n\rListDatabases\x12\x18.tx.ListDatabasesRequest\x1a\x19.tx.ListDatabasesResponse\"\x19\x82\xd3\xe4\x93\x02\x13\x12\x11/api/v1/databases\x12\x61\n\tGetSchema\x12\x14.tx.GetSchemaRequest\x1a\x15.tx.GetSchemaResponse\"\'\x82\xd3\xe4\x93\x02!\x12\x1f/api/v1/databases/{dbid}/schemaB\x1eZ\x1ckwil/api/protobuf/tx/v1;txpbb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'kwil.tx.v1.service_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\034kwil/api/protobuf/tx/v1;txpb'
  _TXSERVICE.methods_by_name['Broadcast']._options = None
  _TXSERVICE.methods_by_name['Broadcast']._serialized_options = b'\202\323\344\223\002\026\"\021/api/v1/broadcast:\001*'
  _TXSERVICE.methods_by_name['EstimatePrice']._options = None
  _TXSERVICE.methods_by_name['EstimatePrice']._serialized_options = b'\202\323\344\223\002\033\"\026/api/v1/estimate_price:\001*'
  _TXSERVICE.methods_by_name['Query']._options = None
  _TXSERVICE.methods_by_name['Query']._serialized_options = b'\202\323\344\223\002\022\"\r/api/v1/query:\001*'
  _TXSERVICE.methods_by_name['GetAccount']._options = None
  _TXSERVICE.methods_by_name['GetAccount']._serialized_options = b'\202\323\344\223\002\034\022\032/api/v1/accounts/{address}'
  _TXSERVICE.methods_by_name['Ping']._options = None
  _TXSERVICE.methods_by_name['Ping']._serialized_options = b'\202\323\344\223\002\016\022\014/api/v1/ping'
  _TXSERVICE.methods_by_name['GetConfig']._options = None
  _TXSERVICE.methods_by_name['GetConfig']._serialized_options = b'\202\323\344\223\002\020\022\016/api/v1/config'
  _TXSERVICE.methods_by_name['ListDatabases']._options = None
  _TXSERVICE.methods_by_name['ListDatabases']._serialized_options = b'\202\323\344\223\002\023\022\021/api/v1/databases'
  _TXSERVICE.methods_by_name['GetSchema']._options = None
  _TXSERVICE.methods_by_name['GetSchema']._serialized_options = b'\202\323\344\223\002!\022\037/api/v1/databases/{dbid}/schema'
  _TXSERVICE._serialized_start=262
  _TXSERVICE._serialized_end=978
# @@protoc_insertion_point(module_scope)



import logging
from typing import Any, Optional

from kwil.types import RPCResponse
from kwil.provider import BaseProvider, AutoProvider
from kwil.exceptions import BadResponseFormat


class RequestManager:
    logger = logging.getLogger("kwil.RequestManager")

    _provider = None

    def __init__(
        self,
        kwil: "Kwil",
        provider: Optional[BaseProvider] = None,
        # middlewares: Optional[List[Tuple[str, Middleware]]] = None,
    ):
        self.kwil = kwil

        if provider is None:
            self.provider = AutoProvider()
        else:
            self.provider = provider

        # if middlewares is None:
        #     self.middlewares = self.default_middlewares()

    @property
    def provider(self) -> BaseProvider:
        return self._provider

    @provider.setter
    def provider(self, provider: BaseProvider) -> None:
        self._provider = provider

    # @staticmethod
    # def default_middlewares() -> List[Tuple[Middleware]]:
    #     return ()

    def _make_request(self, method, params: Any) -> RPCResponse:
        # request_func = self.provider.request_func(self.middlewares)
        request_func = self.provider.request_func()
        self.logger.debug("request %s", method)
        return request_func(method, params)

    @staticmethod
    def formatted_response(response: RPCResponse) -> Any:
        if "error" in response:
            raise ValueError(response["error"])
        elif response.get("result") is not None:
            return response["result"]
        else:
            raise BadResponseFormat("Bad response: {0}".format(response))

    def request_blocking(self, method, params: Any) -> Any:
        resp = self._make_request(method, params)
        return self.formatted_response(resp)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kwil/tx/v1/broadcast.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from kwil.tx.v1 import tx_pb2 as kwil_dot_tx_dot_v1_dot_tx__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1akwil/tx/v1/broadcast.proto\x12\x02tx\x1a\x13kwil/tx/v1/tx.proto\"&\n\x10\x42roadcastRequest\x12\x12\n\x02tx\x18\x01 \x01(\x0b\x32\x06.tx.Tx\"3\n\x11\x42roadcastResponse\x12\x1e\n\x07receipt\x18\x01 \x01(\x0b\x32\r.tx.TxReceiptB\x1eZ\x1ckwil/api/protobuf/tx/v1;txpbb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'kwil.tx.v1.broadcast_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\034kwil/api/protobuf/tx/v1;txpb'
  _BROADCASTREQUEST._serialized_start=55
  _BROADCASTREQUEST._serialized_end=93
  _BROADCASTRESPONSE._serialized_start=95
  _BROADCASTRESPONSE._serialized_end=146
# @@protoc_insertion_point(module_scope)



import logging
from typing import TypeVar, Callable, Any, Optional, Generic, Type

from toolz import pipe

from kwil.types import RPCEndpoint, TReturn
from kwil._utils.method_formatters import (
    get_request_formatters,
    get_response_formatters,
)


TFunc = TypeVar("TFunc", bound=Callable[..., Any])

logger = logging.getLogger(__name__)


def _apply_request_validator(
    method: RPCEndpoint, params: Any, validators: Callable[..., TReturn]
) -> None:
    if validators:
        logger.debug("apply request validator to '%s'", method)
        pipe(params, validators)


class Method(Generic[TFunc]):
    def __init__(
        self,
        rpc_method: RPCEndpoint,
        request_formatters: Optional[Callable[..., TReturn]] = None,
        response_formatters: Optional[Callable[..., TReturn]] = None,
    ):
        self.rpc_method = rpc_method
        self.request_formatters = request_formatters or get_request_formatters
        self.response_formatters = response_formatters or get_response_formatters

    # as Descriptor
    def __get__(
        self, obj: Optional["Module"] = None, obj_type: Optional[Type["Module"]] = None
    ) -> TFunc:
        if obj is None:
            raise TypeError("Methods must be called on an instance of a Module class")
        return obj.caller_fn(obj.kwil, self)

    def method_selector_fn(self) -> Callable[..., Any]:
        raise NotImplementedError("Must be implemented by subclasses")

    def process_params(self, module: "Module", *args: Any) -> Any:
        # to simplify, kwargs are not supported

        method = self.rpc_method

        response_formatters = self.response_formatters(method, module)

        _apply_request_validator(self.rpc_method, args, self.request_formatters(method))

        return (method, args), response_formatters



import json
import os
from typing import Optional, Dict, Any, List, cast

from eth_account.signers.local import LocalAccount

from kwil.kwild import Kwild
from kwil.provider import BaseProvider, GRPCProvider
from kwil.manager import RequestManager as DefaultRequestManager
from kwil.types import (
    TxPayloadType,
    TxParams,
    Nonce,
    TxReceipt,
    DatasetIdentifier,
    DBIdentifier,
    ActionExecution,
)
from kwil._utils.transaction import sign_tx
from kwil._utils.action import encode_action_inputs
from kwil._utils.dataset import generate_dbi
from kwil._utils.signing import load_wallet


class BaseKwil:
    GRPCProvider = GRPCProvider
    RequestManager = DefaultRequestManager

    kwild: Kwild
    web3: None

    @property
    def api(self) -> str:
        from kwil import __version__

        return __version__

    @classmethod
    def generate_dbi(cls, owner: str, db_name: str) -> DBIdentifier:
        return generate_dbi(owner, db_name)


class Kwil(BaseKwil):
    def __init__(
        self,
        provider: Optional[BaseProvider] = None,
        wallet: Optional[LocalAccount] = None,
    ):
        self.manager = self.RequestManager(self, provider)

        self.kwild = Kwild(self)

        if wallet is None:
            self._wallet = load_wallet(os.getenv("KWIL_ETH_PRIVATE_KEY"))
        else:
            self._wallet = wallet

    @property
    def kwil_provider(self) -> BaseProvider:
        return cast(BaseProvider, self.manager.provider)

    @kwil_provider.setter
    def kwil_provider(self, provider: BaseProvider) -> None:
        self.manager.provider = provider

    @property
    def wallet(self) -> LocalAccount:
        return self._wallet

    @wallet.setter
    def wallet(self, wallet: LocalAccount) -> None:
        self._wallet = wallet

    def is_connected(self) -> bool:
        return self.kwil_provider.is_connected()

    def _create_tx(self, payload_type: TxPayloadType, payload: bytes) -> TxParams:
        account_info = self.kwild.get_account(self.wallet.address)

        tx_params = TxParams(
            payloadType=payload_type,
            payload=payload,
            sender=self.wallet.address,
            fee="0",
            # SHOULD be done in result formatter
            nonce=Nonce(int(account_info["nonce"]) + 1),
            # nonce=Nonce(1),
        )

        price = self.kwild.estimate_price(tx_params)
        tx_params["fee"] = price
        tx_params = sign_tx(tx_params, self.wallet)
        return tx_params

    def deploy_database(self, payload: bytes) -> TxReceipt:
        payload_type = TxPayloadType.DEPLOY_DATABASE
        tx_params = self._create_tx(payload_type, payload)
        return self.kwild.broadcast(tx_params)

    def get_database(self, name: str, owner: Optional[str] = None) -> Dict[str, Any]:
        if owner is None:
            owner = self.wallet.address
        db_identifier = generate_dbi(owner, name)
        return self.kwild.get_schema(db_identifier)

    def list_database(self, address: Optional[str] = None) -> List[Dict[str, Any]]:
        if address is None:
            address = self.wallet.address
        return self.kwild.list_database(address)

    def drop_database(self, name: str) -> TxReceipt:
        payload_type = TxPayloadType.DROP_DATABASE
        db_identifier = DatasetIdentifier(owner=self.wallet.address, name=name)
        payload = json.dumps(db_identifier)
        tx_params = self._create_tx(payload_type, payload.encode())
        return self.kwild.broadcast(tx_params)

    def execute_action(
        self,
        db_name: str,
        action_name: str,
        inputs: List[Dict[str, Any]],
    ) -> TxReceipt:
        # TODO: dynamic call action
        db_identifier = generate_dbi(self.wallet.address, db_name)
        encoded_inputs = encode_action_inputs(inputs)
        exec_body = ActionExecution(
            action=action_name, dbID=db_identifier, params=encoded_inputs
        )
        payload_type = TxPayloadType.EXECUTE_ACTION
        payload = json.dumps(exec_body).encode()
        tx_params = self._create_tx(payload_type, payload)
        return self.kwild.broadcast(tx_params)

    def query(self, db_name: str, query: str) -> TxReceipt:
        db_identifier = generate_dbi(self.wallet.address, db_name)
        return self.kwild.query(db_identifier, query)



import grpc

from kwil.types import TxParams, DBIdentifier, HexAddress
from kwil.tx.v1 import (
    ping_pb2,
    query_pb2,
    broadcast_pb2,
    account_pb2,
    config_pb2,
    price_pb2,
    dataset_pb2,
    tx_pb2,
    list_pb2,
    service_pb2_grpc
)

# timeout in seconds
READY_TIMEOUT = 3
REQUEST_TIMEOUT = 2


class Client:
    def __init__(
        self,
        endpoint: str,
        ready_timeout: int = READY_TIMEOUT,
        request_timeout: int = REQUEST_TIMEOUT,
    ):
        self.ready_timeout = ready_timeout
        self.request_timeout = request_timeout
        self.channel = grpc.insecure_channel(endpoint)
        grpc.channel_ready_future(self.channel).result(timeout=self.ready_timeout)
        self.tx_stub = service_pb2_grpc.TxServiceStub(self.channel)

    def close(self):
        self.channel.close()

    def ping(self) -> ping_pb2.PingResponse:
        req = ping_pb2.PingRequest(message="ping")
        return self.tx_stub.Ping(req, timeout=self.request_timeout)

    def query(self, db_id: DBIdentifier, query: str) -> query_pb2.QueryResponse:
        req = query_pb2.QueryRequest(dbid=db_id, query=query)
        return self.tx_stub.Query(req, timeout=self.request_timeout)

    def broadcast(self, tx: TxParams) -> broadcast_pb2.BroadcastResponse:
        gtx = tx_pb2.Tx(
            hash=tx.get("hash"),
            nonce=tx.get("nonce"),
            fee=tx.get("fee"),
            payload_type=tx.get("payloadType"),
            payload=tx.get("payload"),
            signature=tx.get("signature"),
            sender=tx.get("sender"),
        )

        req = broadcast_pb2.BroadcastRequest(tx=gtx)
        return self.tx_stub.Broadcast(req, timeout=self.request_timeout)

    def estimate_price(self, tx: TxParams) -> price_pb2.EstimatePriceResponse:
        gtx = tx_pb2.Tx(
            nonce=tx.get("nonce"),
            fee=tx.get("fee"),
            payload_type=tx.get("payloadType"),
            payload=tx.get("payload"),
        )

        req = price_pb2.EstimatePriceRequest(tx=gtx)
        return self.tx_stub.EstimatePrice(req, timeout=self.request_timeout)

    def get_config(self) -> config_pb2.GetConfigResponse:
        req = config_pb2.GetConfigRequest()
        return self.tx_stub.GetConfig(req, timeout=self.request_timeout)

    def get_account(self, address: HexAddress) -> account_pb2.GetAccountResponse:
        req = account_pb2.GetAccountRequest(address=address)
        return self.tx_stub.GetAccount(req, timeout=self.request_timeout)

    def get_schema(self, db_id: DBIdentifier) -> dataset_pb2.GetSchemaResponse:
        req = dataset_pb2.GetSchemaRequest(dbid=db_id)
        return self.tx_stub.GetSchema(req, timeout=self.request_timeout)

    def list_database(self, owner: HexAddress) -> list_pb2.ListDatabasesResponse:
        req = list_pb2.ListDatabasesRequest(owner=owner)
        return self.tx_stub.ListDatabases(req, timeout=self.request_timeout)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kwil/tx/v1/config.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17kwil/tx/v1/config.proto\x12\x02tx\"\x12\n\x10GetConfigRequest\"\x83\x01\n\x11GetConfigResponse\x12\x1e\n\nchain_code\x18\x01 \x01(\x03R\nchain_code\x12*\n\x10provider_address\x18\x02 \x01(\tR\x10provider_address\x12\"\n\x0cpool_address\x18\x03 \x01(\tR\x0cpool_addressB\x1eZ\x1ckwil/api/protobuf/tx/v1;txpbb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'kwil.tx.v1.config_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\034kwil/api/protobuf/tx/v1;txpb'
  _GETCONFIGREQUEST._serialized_start=31
  _GETCONFIGREQUEST._serialized_end=49
  _GETCONFIGRESPONSE._serialized_start=52
  _GETCONFIGRESPONSE._serialized_end=183
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kwil/tx/v1/list.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15kwil/tx/v1/list.proto\x12\x02tx\"%\n\x14ListDatabasesRequest\x12\r\n\x05owner\x18\x01 \x01(\t\"*\n\x15ListDatabasesResponse\x12\x11\n\tdatabases\x18\x01 \x03(\tB\x1eZ\x1ckwil/api/protobuf/tx/v1;txpbb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'kwil.tx.v1.list_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\034kwil/api/protobuf/tx/v1;txpb'
  _LISTDATABASESREQUEST._serialized_start=29
  _LISTDATABASESREQUEST._serialized_end=66
  _LISTDATABASESRESPONSE._serialized_start=68
  _LISTDATABASESRESPONSE._serialized_end=110
# @@protoc_insertion_point(module_scope)



from kwil.types import RPCEndpoint


class GRPC:
    kwil_getSchema = RPCEndpoint("kwil_getSchema")
    kwil_getConfig = RPCEndpoint("kwil_getConfig")
    kwil_getAccount = RPCEndpoint("kwil_getAccount")
    kwil_listDatabase = RPCEndpoint("kwil_listDatabase")
    kwil_ping = RPCEndpoint("kwil_ping")
    kwil_getBlock = RPCEndpoint("kwil_getBlock")
    kwil_broadcast = RPCEndpoint("kwil_broadcast")
    kwil_estimatePrice = RPCEndpoint("kwil_estimatePrice")
    kwil_query = RPCEndpoint("kwil_query")



import logging
from typing import Callable, Any

from toolz import curry, pipe

from kwil.method import Method
from kwil.types import RPCEndpoint, TReturn

logger = logging.getLogger(__name__)


@curry
def _apply_result_formatters(
    method: RPCEndpoint, result_formatters: Callable[..., TReturn], result: Any
) -> Any:
    if result_formatters:
        logger.debug("apply result formatter to '%s'", method)
        formatted_result = pipe(result, result_formatters)
        return formatted_result
    else:
        return result


class Module:
    @curry
    def caller_fn(
        self, kwil: "Kwil", method: Method[Callable[..., Any]]
    ) -> Callable[..., Any]:
        def caller(*args: Any) -> Any:
            (method_str, args), response_formatters = method.process_params(self, *args)
            result = kwil.manager.request_blocking(method_str, args)
            return _apply_result_formatters(
                method.rpc_method, response_formatters, result
            )

        return caller



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from kwil.tx.v1 import account_pb2 as kwil_dot_tx_dot_v1_dot_account__pb2
from kwil.tx.v1 import broadcast_pb2 as kwil_dot_tx_dot_v1_dot_broadcast__pb2
from kwil.tx.v1 import config_pb2 as kwil_dot_tx_dot_v1_dot_config__pb2
from kwil.tx.v1 import dataset_pb2 as kwil_dot_tx_dot_v1_dot_dataset__pb2
from kwil.tx.v1 import list_pb2 as kwil_dot_tx_dot_v1_dot_list__pb2
from kwil.tx.v1 import ping_pb2 as kwil_dot_tx_dot_v1_dot_ping__pb2
from kwil.tx.v1 import price_pb2 as kwil_dot_tx_dot_v1_dot_price__pb2
from kwil.tx.v1 import query_pb2 as kwil_dot_tx_dot_v1_dot_query__pb2


class TxServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Broadcast = channel.unary_unary(
                '/tx.TxService/Broadcast',
                request_serializer=kwil_dot_tx_dot_v1_dot_broadcast__pb2.BroadcastRequest.SerializeToString,
                response_deserializer=kwil_dot_tx_dot_v1_dot_broadcast__pb2.BroadcastResponse.FromString,
                )
        self.EstimatePrice = channel.unary_unary(
                '/tx.TxService/EstimatePrice',
                request_serializer=kwil_dot_tx_dot_v1_dot_price__pb2.EstimatePriceRequest.SerializeToString,
                response_deserializer=kwil_dot_tx_dot_v1_dot_price__pb2.EstimatePriceResponse.FromString,
                )
        self.Query = channel.unary_unary(
                '/tx.TxService/Query',
                request_serializer=kwil_dot_tx_dot_v1_dot_query__pb2.QueryRequest.SerializeToString,
                response_deserializer=kwil_dot_tx_dot_v1_dot_query__pb2.QueryResponse.FromString,
                )
        self.GetAccount = channel.unary_unary(
                '/tx.TxService/GetAccount',
                request_serializer=kwil_dot_tx_dot_v1_dot_account__pb2.GetAccountRequest.SerializeToString,
                response_deserializer=kwil_dot_tx_dot_v1_dot_account__pb2.GetAccountResponse.FromString,
                )
        self.Ping = channel.unary_unary(
                '/tx.TxService/Ping',
                request_serializer=kwil_dot_tx_dot_v1_dot_ping__pb2.PingRequest.SerializeToString,
                response_deserializer=kwil_dot_tx_dot_v1_dot_ping__pb2.PingResponse.FromString,
                )
        self.GetConfig = channel.unary_unary(
                '/tx.TxService/GetConfig',
                request_serializer=kwil_dot_tx_dot_v1_dot_config__pb2.GetConfigRequest.SerializeToString,
                response_deserializer=kwil_dot_tx_dot_v1_dot_config__pb2.GetConfigResponse.FromString,
                )
        self.ListDatabases = channel.unary_unary(
                '/tx.TxService/ListDatabases',
                request_serializer=kwil_dot_tx_dot_v1_dot_list__pb2.ListDatabasesRequest.SerializeToString,
                response_deserializer=kwil_dot_tx_dot_v1_dot_list__pb2.ListDatabasesResponse.FromString,
                )
        self.GetSchema = channel.unary_unary(
                '/tx.TxService/GetSchema',
                request_serializer=kwil_dot_tx_dot_v1_dot_dataset__pb2.GetSchemaRequest.SerializeToString,
                response_deserializer=kwil_dot_tx_dot_v1_dot_dataset__pb2.GetSchemaResponse.FromString,
                )


class TxServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Broadcast(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EstimatePrice(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Query(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAccount(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Ping(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListDatabases(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSchema(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TxServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Broadcast': grpc.unary_unary_rpc_method_handler(
                    servicer.Broadcast,
                    request_deserializer=kwil_dot_tx_dot_v1_dot_broadcast__pb2.BroadcastRequest.FromString,
                    response_serializer=kwil_dot_tx_dot_v1_dot_broadcast__pb2.BroadcastResponse.SerializeToString,
            ),
            'EstimatePrice': grpc.unary_unary_rpc_method_handler(
                    servicer.EstimatePrice,
                    request_deserializer=kwil_dot_tx_dot_v1_dot_price__pb2.EstimatePriceRequest.FromString,
                    response_serializer=kwil_dot_tx_dot_v1_dot_price__pb2.EstimatePriceResponse.SerializeToString,
            ),
            'Query': grpc.unary_unary_rpc_method_handler(
                    servicer.Query,
                    request_deserializer=kwil_dot_tx_dot_v1_dot_query__pb2.QueryRequest.FromString,
                    response_serializer=kwil_dot_tx_dot_v1_dot_query__pb2.QueryResponse.SerializeToString,
            ),
            'GetAccount': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAccount,
                    request_deserializer=kwil_dot_tx_dot_v1_dot_account__pb2.GetAccountRequest.FromString,
                    response_serializer=kwil_dot_tx_dot_v1_dot_account__pb2.GetAccountResponse.SerializeToString,
            ),
            'Ping': grpc.unary_unary_rpc_method_handler(
                    servicer.Ping,
                    request_deserializer=kwil_dot_tx_dot_v1_dot_ping__pb2.PingRequest.FromString,
                    response_serializer=kwil_dot_tx_dot_v1_dot_ping__pb2.PingResponse.SerializeToString,
            ),
            'GetConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.GetConfig,
                    request_deserializer=kwil_dot_tx_dot_v1_dot_config__pb2.GetConfigRequest.FromString,
                    response_serializer=kwil_dot_tx_dot_v1_dot_config__pb2.GetConfigResponse.SerializeToString,
            ),
            'ListDatabases': grpc.unary_unary_rpc_method_handler(
                    servicer.ListDatabases,
                    request_deserializer=kwil_dot_tx_dot_v1_dot_list__pb2.ListDatabasesRequest.FromString,
                    response_serializer=kwil_dot_tx_dot_v1_dot_list__pb2.ListDatabasesResponse.SerializeToString,
            ),
            'GetSchema': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSchema,
                    request_deserializer=kwil_dot_tx_dot_v1_dot_dataset__pb2.GetSchemaRequest.FromString,
                    response_serializer=kwil_dot_tx_dot_v1_dot_dataset__pb2.GetSchemaResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'tx.TxService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class TxService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Broadcast(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/tx.TxService/Broadcast',
            kwil_dot_tx_dot_v1_dot_broadcast__pb2.BroadcastRequest.SerializeToString,
            kwil_dot_tx_dot_v1_dot_broadcast__pb2.BroadcastResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def EstimatePrice(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/tx.TxService/EstimatePrice',
            kwil_dot_tx_dot_v1_dot_price__pb2.EstimatePriceRequest.SerializeToString,
            kwil_dot_tx_dot_v1_dot_price__pb2.EstimatePriceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Query(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/tx.TxService/Query',
            kwil_dot_tx_dot_v1_dot_query__pb2.QueryRequest.SerializeToString,
            kwil_dot_tx_dot_v1_dot_query__pb2.QueryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetAccount(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/tx.TxService/GetAccount',
            kwil_dot_tx_dot_v1_dot_account__pb2.GetAccountRequest.SerializeToString,
            kwil_dot_tx_dot_v1_dot_account__pb2.GetAccountResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Ping(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/tx.TxService/Ping',
            kwil_dot_tx_dot_v1_dot_ping__pb2.PingRequest.SerializeToString,
            kwil_dot_tx_dot_v1_dot_ping__pb2.PingResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/tx.TxService/GetConfig',
            kwil_dot_tx_dot_v1_dot_config__pb2.GetConfigRequest.SerializeToString,
            kwil_dot_tx_dot_v1_dot_config__pb2.GetConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListDatabases(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/tx.TxService/ListDatabases',
            kwil_dot_tx_dot_v1_dot_list__pb2.ListDatabasesRequest.SerializeToString,
            kwil_dot_tx_dot_v1_dot_list__pb2.ListDatabasesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetSchema(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/tx.TxService/GetSchema',
            kwil_dot_tx_dot_v1_dot_dataset__pb2.GetSchemaRequest.SerializeToString,
            kwil_dot_tx_dot_v1_dot_dataset__pb2.GetSchemaResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from typing import Optional, Union

from eth_account import Account
from eth_utils import to_bytes, keccak as eth_utils_keccak
from hexbytes import (
    HexBytes,
)

from kwil_typing import HexStr
from kwil.types import Signature, SignatureType


def keccak(
    primitive: Union[bytes, int, bool, None] = None,
    text: Optional[str] = None,
    hexstr: Optional[HexStr] = None,
) -> HexBytes:
    if isinstance(primitive, (bytes, int, type(None))):
        input_bytes = to_bytes(primitive, hexstr=hexstr, text=text)
        return HexBytes(eth_utils_keccak(input_bytes))

    raise TypeError(
        f"You called keccak with first arg {primitive!r} and keywords "
        f"{{'text': {text!r}, 'hexstr': {hexstr!r}}}. You must call it with "
        "one of these approaches: keccak(text='txt'), keccak(hexstr='0x747874'), "
        "keccak(b'\\x74\\x78\\x74'), or keccak(0x747874)."
    )


def sign(data: bytes, private_key: HexStr) -> Signature:
    signature_type = SignatureType.PK_SECP256K1_UNCOMPRESSED
    message_hash_bytes = keccak(data)

    # message_hash_hex = message_hash_bytes.hex()
    # signable_message = messages.encode_defunct(message_hash_bytes)
    # signature_bytes = Account.sign_message(signable_message, private_key).signature
    # print("signature_bytes 000", [b for b in signature_bytes])

    signature_bytes = Account.signHash(message_hash_bytes, private_key).signature
    signature_byte_array = bytearray()
    signature_byte_array.extend(signature_bytes[:-1])

    # set V=0, to keep aligned with Go IMPL
    # https://github.com/ethereum/go-ethereum/blob/v1.10.21/crypto/secp256k1/secp256.go#L97
    signature_byte_array.append(signature_bytes[-1] - 27)
    return Signature(
        signature_bytes=bytes(signature_byte_array), signature_type=signature_type
    )



from typing import (
    Callable,
    Sequence,
    Any,
)

import toolz

from kwil.types import RPCResponse, Middleware, RPCEndpoint
from kwil._utils.validation import request_validators


def combine_middlewares(
    middlewares: Sequence[Middleware],
    provider_request_fn: Callable[[RPCEndpoint, Any], Any],
) -> Callable[..., RPCResponse]:
    """
    Returns a callable function which will call the provider_request_fn
    wrapped with all the middlewares.
    Middlewares are applied in reverse order, so the first middleware in
    the list will be the outermost.
    """
    return toolz.reduce(
        lambda request_fn, middleware: middleware(request_fn),
        reversed(middlewares),
        provider_request_fn,
    )


validation_middlewares = request_validators



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kwil/tx/v1/query.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16kwil/tx/v1/query.proto\x12\x02tx\"+\n\x0cQueryRequest\x12\x0c\n\x04\x64\x62id\x18\x01 \x01(\t\x12\r\n\x05query\x18\x02 \x01(\t\"\x1f\n\rQueryResponse\x12\x0e\n\x06result\x18\x01 \x01(\x0c\x42\x1eZ\x1ckwil/api/protobuf/tx/v1;txpbb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'kwil.tx.v1.query_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\034kwil/api/protobuf/tx/v1;txpb'
  _QUERYREQUEST._serialized_start=30
  _QUERYREQUEST._serialized_end=73
  _QUERYRESPONSE._serialized_start=75
  _QUERYRESPONSE._serialized_end=106
# @@protoc_insertion_point(module_scope)



class BadResponseFormat(ValueError, KeyError):
    # Inherits from KeyError and ValueError for backwards compatibility
    """
    Raised when the response format is incorrect.
    """
    pass


class InvalidAddress(ValueError):
    """
    The supplied address is invalid.
    """

    pass


class CannotHandleRequest(ValueError):
    """
    The provider cannot handle the request.
    """

    pass



import logging
from typing import Callable, List

from kwil._utils.rpcs import GRPC
from kwil.method import Method
from kwil.module import Module
from kwil.types import (
    DBIdentifier,
    HexAddress,
    TxParams,
    TxReceipt,
    AccountInfo,
    DBSchema,
    ServiceConfig,
)


class BaseKwild(Module):
    """Kwil Chain base class"""

    logger = logging.getLogger("kwil.BaseKwild")

    def __init__(self, kwil):
        self.kwil = kwil
        self.cfg = self.get_config()
        self.logger.debug("KwilChain config: %s", self.cfg)

    # def get_block(self, block_id):
    #     return self.kwil.manager.request_blocking(GRPC.kwil_getBlock, [block_id])

    ping: Method[Callable[[], bool]] = Method(GRPC.kwil_ping)
    get_config: Method[Callable[[], ServiceConfig]] = Method(GRPC.kwil_getConfig)


class Kwild(BaseKwild):
    """Kwil Chain class"""

    get_schema: Method[Callable[[DBIdentifier], DBSchema]] = Method(GRPC.kwil_getSchema)
    list_database: Method[Callable[[HexAddress], List[DBSchema]]] = Method(
        GRPC.kwil_listDatabase
    )
    get_account: Method[Callable[[HexAddress], AccountInfo]] = Method(
        GRPC.kwil_getAccount
    )
    broadcast: Method[Callable[[TxParams], TxReceipt]] = Method(GRPC.kwil_broadcast)
    estimate_price: Method[Callable[[TxParams], str]] = Method(GRPC.kwil_estimatePrice)
    query: Method[Callable[[DBIdentifier, str], TxReceipt]] = Method(GRPC.kwil_query)



from typing import (
    Any,
    List,
    Dict,
    NewType,
    TypeVar,
    TypedDict,
    Callable,
)
from enum import IntEnum

from kwil_typing import (
    DBIdentifier,
    HexAddress,
)


class RPCResponse(TypedDict, total=False):
    error: Any
    result: Any


URI = NewType("URI", str)
RPCEndpoint = NewType("RPCEndpoint", str)
Nonce = NewType("Nonce", int)
TReturn = TypeVar("TReturn")
TParams = TypeVar("TParams")
Middleware = Callable[[RPCEndpoint, Any], RPCResponse]


class DatasetIdentifier(TypedDict, total=True):
    owner: str
    name: str


class TxPayloadType(IntEnum):
    INVALID_PAYLOAD_TYPE = 100
    DEPLOY_DATABASE = 101
    MODIFY_DATABASE = 102
    DROP_DATABASE = 103
    EXECUTE_ACTION = 104
    END_PAYLOAD_TYPE = 105

    def __str__(self):
        return f"{self.name}"


class SignatureType(IntEnum):
    SIGNATURE_TYPE_INVALID = 0
    PK_SECP256K1_UNCOMPRESSED = 1
    ACCOUNT_SECP256K1_UNCOMPRESSED = 2
    END_SIGNATURE_TYPE = 3

    def __str__(self):
        return f"{self.name}"


class ColumnType(IntEnum):
    INVALID_COLUMN_TYPE = 100
    NULL = 101
    TEXT = 102
    INT = 103
    END_COLUMN_TYPE = 104

    def __str__(self):
        return f"{self.name}"


class AttributeType(IntEnum):
    INVALID_ATTRIBUTE_TYPE = 0
    PRIMARY_KEY = 1
    UNIQUE = 2
    NOT_NULL = 3
    DEFAULT = 4
    MIN = 5
    MAX = 6
    MIN_LENGTH = 7
    MAX_LENGTH = 8
    END_ATTRIBUTE_TYPE = 4

    def __str__(self):
        return f"{self.name}"


class IndexType(IntEnum):
    INVALID_INDEX_TYPE = 100
    BTREE = 101
    UNIQUE_BTREE = 102
    END_INDEX_TYPE = 103

    def __str__(self):
        return f"{self.name}"


class Signature(TypedDict, total=True):
    signature_bytes: bytes
    signature_type: SignatureType


class TxParams(TypedDict, total=False):
    hash: bytes
    payloadType: TxPayloadType
    payload: bytes
    fee: str
    nonce: Nonce
    signature: Signature
    sender: str


class TxReceipt(TypedDict, total=False):
    hash: bytes
    fee: str
    body: Dict[str, Any]
    result: Dict[str, Any]


class AccountInfo(TypedDict, total=False):
    address: HexAddress
    nonce: Nonce
    balance: str


class AttributeSchema(TypedDict, total=False):
    type: AttributeType
    value: bytes


class ColumnSchema(TypedDict, total=False):
    name: str
    type: ColumnType
    attributes: List[AttributeSchema]


class IndexSchema(TypedDict, total=False):
    name: str
    columns: List[str]
    type: IndexType


class TableSchema(TypedDict, total=False):
    name: str
    columns: List[ColumnSchema]
    indexes: List[IndexSchema]


class ActionSchema(TypedDict, total=False):
    name: str
    public: bool
    inputs: List[str]
    statements: List[str]


class DBSchema(TypedDict, total=False):
    owner: HexAddress
    name: str
    tables: List[TableSchema]
    actions: List[ActionSchema]


class ServiceConfig(TypedDict, total=False):
    chainCode: int
    providerAddress: HexAddress
    poolAddress: HexAddress


class ActionExecution(TypedDict, total=False):
    action: str
    dbID: DBIdentifier
    params: List[Dict[str, Any]]


class ActionParamDataType(IntEnum):
    INVALID_DATA_TYPE = 100
    NULL = 101
    TEXT = 102
    INT = 103
    END_DATA_TYPE = 104

    def __str__(self):
        return f"{self.name}"


class ActionParamValue(TypedDict, total=False):
    value: Any
    dataType: ActionParamDataType
    bytes: str



def camel_to_snake(name: str) -> str:
    """Converts a camelCase string to snake_case.

    Args:
        name (str): The camelCase string to convert.

    Returns:
        str: The snake_case string.
    """
    result = ""
    for i, c in enumerate(name):
        if c.isupper():
            if i > 0:
                result += "_"
            result += c.lower()
        else:
            result += c
    return result


def rpc_to_grpc_method_name(rpc_method_name: str) -> str:
    """Converts an RPC method name to a gRPC method name.

    Args:
        rpc_method_name (str): The RPC method name.

    Returns:
        str: The gRPC method name.
    """
    return camel_to_snake(rpc_method_name.split("_")[-1])



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kwil/tx/v1/dataset.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18kwil/tx/v1/dataset.proto\x12\x02tx\" \n\x10GetSchemaRequest\x12\x0c\n\x04\x64\x62id\x18\x01 \x01(\t\"1\n\x11GetSchemaResponse\x12\x1c\n\x07\x64\x61taset\x18\x01 \x01(\x0b\x32\x0b.tx.Dataset\"^\n\x07\x44\x61taset\x12\r\n\x05owner\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x19\n\x06tables\x18\x03 \x03(\x0b\x32\t.tx.Table\x12\x1b\n\x07\x61\x63tions\x18\x04 \x03(\x0b\x32\n.tx.Action\"N\n\x05Table\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1b\n\x07\x63olumns\x18\x02 \x03(\x0b\x32\n.tx.Column\x12\x1a\n\x07indexes\x18\x03 \x03(\x0b\x32\t.tx.Index\"U\n\x06\x43olumn\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1a\n\x04type\x18\x02 \x01(\x0e\x32\x0c.tx.DataType\x12!\n\nattributes\x18\x03 \x03(\x0b\x32\r.tx.Attribute\";\n\tAttribute\x12\x1f\n\x04type\x18\x01 \x01(\x0e\x32\x11.tx.AttributeType\x12\r\n\x05value\x18\x02 \x01(\x0c\"C\n\x05Index\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07\x63olumns\x18\x02 \x03(\t\x12\x1b\n\x04type\x18\x03 \x01(\x0e\x32\r.tx.IndexType\"J\n\x06\x41\x63tion\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06public\x18\x02 \x01(\x08\x12\x0e\n\x06inputs\x18\x03 \x03(\t\x12\x12\n\nstatements\x18\x04 \x03(\t*^\n\x08\x44\x61taType\x12\x15\n\x11INVALID_DATA_TYPE\x10\x00\x12\x12\n\x0eNULL_DATA_TYPE\x10\x01\x12\x14\n\x10STRING_DATA_TYPE\x10\x02\x12\x11\n\rINT_DATA_TYPE\x10\x03*\x90\x01\n\rAttributeType\x12\x15\n\x11INVALID_ATTRIBUTE\x10\x00\x12\x0f\n\x0bPRIMARY_KEY\x10\x01\x12\n\n\x06UNIQUE\x10\x02\x12\x0c\n\x08NOT_NULL\x10\x03\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x04\x12\x07\n\x03MIN\x10\x05\x12\x07\n\x03MAX\x10\x06\x12\x0e\n\nMIN_LENGTH\x10\x07\x12\x0e\n\nMAX_LENGTH\x10\x08*V\n\tIndexType\x12\x16\n\x12INVALID_INDEX_TYPE\x10\x00\x12\x14\n\x10\x42TREE_INDEX_TYPE\x10\x01\x12\x1b\n\x17UNIQUE_BTREE_INDEX_TYPE\x10\x02\x42\x1eZ\x1ckwil/api/protobuf/tx/v1;txpbb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'kwil.tx.v1.dataset_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\034kwil/api/protobuf/tx/v1;txpb'
  _DATATYPE._serialized_start=586
  _DATATYPE._serialized_end=680
  _ATTRIBUTETYPE._serialized_start=683
  _ATTRIBUTETYPE._serialized_end=827
  _INDEXTYPE._serialized_start=829
  _INDEXTYPE._serialized_end=915
  _GETSCHEMAREQUEST._serialized_start=32
  _GETSCHEMAREQUEST._serialized_end=64
  _GETSCHEMARESPONSE._serialized_start=66
  _GETSCHEMARESPONSE._serialized_end=115
  _DATASET._serialized_start=117
  _DATASET._serialized_end=211
  _TABLE._serialized_start=213
  _TABLE._serialized_end=291
  _COLUMN._serialized_start=293
  _COLUMN._serialized_end=378
  _ATTRIBUTE._serialized_start=380
  _ATTRIBUTE._serialized_end=439
  _INDEX._serialized_start=441
  _INDEX._serialized_end=508
  _ACTION._serialized_start=510
  _ACTION._serialized_end=584
# @@protoc_insertion_point(module_scope)



from typing import Any, Callable, List

from toolz import curry


@curry
def apply_validator_at_index(
    fn: Callable[..., Any], index: int, value: List[Any]
) -> Any:
    if index >= len(value):
        raise IndexError(
            "Not enough values to apply validator. {0}<{1}".format(len(value), index)
        )
    for i, item in enumerate(value):
        if i == index:
            fn(item)

    return value



import logging
from typing import Any, Optional, Union

import grpc

from kwil.types import RPCEndpoint, RPCResponse, URI
from kwil.provider.base import ProtoBaseProvider
from kwil._utils.request import grpc_request, get_default_grpc_endpoint


class GRPCProvider(ProtoBaseProvider):
    logger = logging.getLogger("kwil.provider.GRPCProvider")

    endpoint_uri = None

    def __init__(self, endpoint_uri: Optional[Union[URI, str]] = None) -> None:
        if endpoint_uri is None:
            self.endpoint_uri = get_default_grpc_endpoint()
        else:
            self.endpoint_uri = URI(endpoint_uri)

        super().__init__()

    def __str__(self) -> str:
        return "GRPC connection {0}".format(self.endpoint_uri)

    def make_request(self, method: RPCEndpoint, params: Any) -> RPCResponse:
        # make request here
        self.logger.debug(
            "Making GRPC request. URI: %s, Method: %s", self.endpoint_uri, method
        )

        try:
            raw_response = grpc_request(self.endpoint_uri, method, params)
        except grpc.RpcError as e:
            self.logger.debug(
                "Call GRPC error. URI: %s, Method: %s, Error: %s",
                self.endpoint_uri,
                method,
                e,
            )
            return RPCResponse(error=e)
        else:
            response = self.decode_rpc_request(raw_response)
            self.logger.debug(
                "Got GRPC response. URI: %s, Method: %s, Response: %s",
                self.endpoint_uri,
                method,
                response,
            )
            return RPCResponse(result=response)



import sys
import pkg_resources

from kwil.main import Kwil
from kwil.provider import GRPCProvider

if sys.version_info < (3, 9):
    raise EnvironmentError("Python 3.9 or above is required. ")

__version__ = pkg_resources.get_distribution("kwil").version

__all__ = [
    "__version__",
    "Kwil",
    "GRPCProvider",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kwil/tx/v1/account.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18kwil/tx/v1/account.proto\x12\x02tx\":\n\x07\x41\x63\x63ount\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x0f\n\x07\x62\x61lance\x18\x02 \x01(\t\x12\r\n\x05nonce\x18\x03 \x01(\x03\"$\n\x11GetAccountRequest\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\"2\n\x12GetAccountResponse\x12\x1c\n\x07\x61\x63\x63ount\x18\x01 \x01(\x0b\x32\x0b.tx.AccountB\x1eZ\x1ckwil/api/protobuf/tx/v1;txpbb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'kwil.tx.v1.account_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\034kwil/api/protobuf/tx/v1;txpb'
  _ACCOUNT._serialized_start=32
  _ACCOUNT._serialized_end=90
  _GETACCOUNTREQUEST._serialized_start=92
  _GETACCOUNTREQUEST._serialized_end=128
  _GETACCOUNTRESPONSE._serialized_start=130
  _GETACCOUNTRESPONSE._serialized_end=180
# @@protoc_insertion_point(module_scope)



from typing import List, Dict, Any
import base64

from kwil.types import ActionParamValue, ActionParamDataType


action_param_text_byte = ActionParamDataType.TEXT.to_bytes(1, byteorder="little")
action_param_int_byte = ActionParamDataType.INT.to_bytes(1, byteorder="little")


def encode_param(_type: bytes, _value: bytes) -> str:
    # encode type and value as base64 string
    bs = bytearray()
    bs.extend(_type)
    bs.extend(_value)
    encoded = base64.b64encode(bs).decode("utf-8")
    return encoded


def _encode_action_value(value: Any) -> ActionParamValue:
    if isinstance(value, str):
        encoded = encode_param(action_param_text_byte, value.encode())
        return ActionParamValue(
            value=value,
            dataType=ActionParamDataType.TEXT,
            bytes=encoded,
        )
    elif isinstance(value, int):
        encoded = encode_param(
            action_param_int_byte, value.to_bytes(8, byteorder="little")
        )
        return ActionParamValue(
            value=value,
            dataType=ActionParamDataType.INT,
            bytes=encoded,
        )
    else:
        raise ValueError(f"Unsupported type {type(value)}")


def encode_action_inputs(inputs: List[Dict[str, Any]]) -> List[Dict[str, str]]:
    encoded_inputs = []
    for input in inputs:
        encoded_input = {}
        for key, value in input.items():
            encoded_value = _encode_action_value(value)
            encoded_input[key] = encoded_value["bytes"]
        encoded_inputs.append(encoded_input)
    return encoded_inputs



import hashlib

from kwil.types import DBIdentifier


def generate_dbi(owner: str, name: str) -> DBIdentifier:
    """Generate a database identifier."""
    bs = bytearray()
    bs.extend(name.lower().encode())
    bs.extend(owner.lower().encode())
    formatted = "x{0}".format(hashlib.sha224(bs).hexdigest())
    return DBIdentifier(formatted)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kwil/tx/v1/tx.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13kwil/tx/v1/tx.proto\x12\x02tx\"\x95\x01\n\x02Tx\x12\x0c\n\x04hash\x18\x01 \x01(\x0c\x12\"\n\x0cpayload_type\x18\x02 \x01(\x05R\x0cpayload_type\x12\x0f\n\x07payload\x18\x03 \x01(\x0c\x12\r\n\x05nonce\x18\x04 \x01(\x03\x12 \n\tsignature\x18\x05 \x01(\x0b\x32\r.tx.Signature\x12\x0b\n\x03\x66\x65\x65\x18\x06 \x01(\t\x12\x0e\n\x06sender\x18\x07 \x01(\t\"p\n\tSignature\x12(\n\x0fsignature_bytes\x18\x01 \x01(\x0cR\x0fsignature_bytes\x12\x39\n\x0esignature_type\x18\x02 \x01(\x0e\x32\x11.tx.SignatureTypeR\x0esignature_type\"7\n\tTxReceipt\x12\x0f\n\x07tx_hash\x18\x01 \x01(\x0c\x12\x0b\n\x03\x66\x65\x65\x18\x02 \x01(\t\x12\x0c\n\x04\x62ody\x18\x03 \x01(\x0c*\x85\x01\n\rSignatureType\x12\x19\n\x15INVALID_SINATURE_TYPE\x10\x00\x12\x1d\n\x19PK_SECP256K1_UNCOMPRESSED\x10\x01\x12\"\n\x1e\x41\x43\x43OUNT_SECP256K1_UNCOMPRESSED\x10\x02\x12\x16\n\x12\x45ND_SIGNATURE_TYPE\x10\x03\x42\x1eZ\x1ckwil/api/protobuf/tx/v1;txpbb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'kwil.tx.v1.tx_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\034kwil/api/protobuf/tx/v1;txpb'
  _SIGNATURETYPE._serialized_start=351
  _SIGNATURETYPE._serialized_end=484
  _TX._serialized_start=28
  _TX._serialized_end=177
  _SIGNATURE._serialized_start=179
  _SIGNATURE._serialized_end=291
  _TXRECEIPT._serialized_start=293
  _TXRECEIPT._serialized_end=348
# @@protoc_insertion_point(module_scope)



import base64
import json
from typing import Dict, Any, Callable

from toolz import curry, compose, identity

from kwil.types import RPCEndpoint
from kwil._utils.rpcs import GRPC
from kwil._utils.validation import request_validators


def get_request_formatters(method_name: RPCEndpoint) -> Callable[..., Any]:
    formatter_maps = (request_validators,)

    return compose(*[fm[method_name] for fm in formatter_maps if method_name in fm])


@curry
def extract_result(key: Any, response: Dict[str, Any]) -> Any:
    return response[key]


def to_bool_if_message_match(message) -> Callable[..., bool]:
    def if_match(response: Any) -> bool:
        return response.message == message

    return if_match


def ping_response_formatter(response: Dict[Any, Any]) -> str:
    return response.get("message", "shoot")


def tx_receipt_decode(response: Dict[str, Any]) -> Dict[str, Any]:
    response = response["receipt"]
    body = response["body"]
    del response["body"]
    # TODO: make server return a standard response, not ""
    if body == "":
        response["result"] = {}
        return response
    result = json.loads(base64.b64decode(body))
    response["result"] = result
    return response


def query_decode(response: Dict[str, Any]) -> Dict[str, Any]:
    base64_response = response["result"]
    result = json.loads(base64.b64decode(base64_response))
    response["result"] = result
    return response


response_formatters: Dict[str, Callable[[Any], Any]] = {
    GRPC.kwil_ping: ping_response_formatter,
    GRPC.kwil_getSchema: extract_result("dataset"),
    GRPC.kwil_getConfig: identity,
    GRPC.kwil_getAccount: extract_result("account"),
    GRPC.kwil_estimatePrice: extract_result("price"),
    GRPC.kwil_broadcast: tx_receipt_decode,
    GRPC.kwil_listDatabase: extract_result("databases"),
    GRPC.kwil_query: query_decode,
}


def get_response_formatters(
    method_name: RPCEndpoint, module: "Module"
) -> Callable[..., Any]:
    formatter_maps = (response_formatters,)

    return compose(
        *[
            formatter_map[method_name]
            for formatter_map in formatter_maps
            if method_name in formatter_map
        ]
    )



import logging
from typing import Callable, Any, Dict, Tuple

from google.protobuf.json_format import MessageToDict
from google.protobuf import message as _message

from kwil.types import RPCEndpoint, RPCResponse, Middleware
from kwil._utils.rpcs import GRPC


class BaseProvider:
    """Handling middlewares"""

    _middlewares: Tuple[Middleware, ...] = ()
    # cache by middlewares, (all_middlewares, request_func)
    _request_func_cache: Tuple[Tuple[Middleware, ...], Callable[..., RPCResponse]] = (
        None,
        None,
    )

    @property
    def middlewares(self) -> Tuple[Middleware, ...]:
        return self._middlewares

    @middlewares.setter
    def middlewares(self, middlewares: Tuple[Middleware, ...]) -> None:
        self._middlewares = middlewares

    def request_func(
        self,
        # middlewares: List[Middleware]
    ) -> Callable[..., RPCResponse]:
        # all_middlewares: Tuple[Middleware] = tuple(middlewares) +
        #   tuple(self.middlewares)
        # cache_key = self._request_func_cache[0]
        # if cache_key is None or cache_key != all_middlewares:
        #     self._request_func_cache = (
        #         all_middlewares,
        #         combine_middlewares(all_middlewares, self.make_request))
        # return self._request_func_cache[1]
        return self.make_request

    def make_request(self, method: RPCEndpoint, params: Any) -> RPCResponse:
        raise NotImplementedError("Provider must implement this method")

    def is_connected(self) -> bool:
        raise NotImplementedError("Provider must implement this method")


class ProtoBaseProvider(BaseProvider):
    """Handling serialization"""

    def __str__(self) -> None:
        pass

    def encode_rpc_request(self, method, params) -> bytes:
        # encode here
        pass

    def decode_rpc_request(self, raw_response: _message) -> Dict[Any, Any]:
        return MessageToDict(
            raw_response,
            # descriptor_pool=_descriptor_pool.Default(),
            including_default_value_fields=True,
        )

    def is_connected(self) -> bool:
        try:
            response = self.make_request(GRPC.kwil_ping, "")
        except Exception as ex:
            logging.exception("caught error", ex)
            return False

        assert "error" not in response
        return True
